import java.util.ArrayList;


/**
 * A simple class to test our various shapes.
 *
 * @author bsnider@georgefox.edu
 */
public class ShapeTest
{
    /**
     * The main entry point of the program.
     *
     * @param args Command-line arguments to this program (unused)
     */
    public static void main(String[] args)
    {
        // we cannot create new instances of abstract class Shape,
        // because some methods are abstract and have no implementation
        // Shape x = new Shape();  // won't compile

        // we can, however, create new instances of specific types of shapes
        // as these subclasses are concrete and fully implemented
        Circle circle = new Circle(3.5);
        Rectangle rectangle = new Rectangle(1, 4);
        Square square = new Square(2);

        // note that we cannot somehow create a new Square with different
        // length and width attributes, even though its superclass Rectangle
        // supports this behavior
        // Square s = new Square(3, 3);  // won't compile

        // despite not being able to create new instances of abstract class
        // Shape, we can still generally refer to any Shape subclass because
        // of the object-oriented concept of substitutability, which says that
        // a subclass "is-a" superclass.
        // we'll test this out by adding our subclass instances to a
        // collection that holds Shape objects
        ArrayList<Shape> shapes = new ArrayList<>();
        shapes.add(circle);
        shapes.add(rectangle);
        shapes.add(square);

        // we can print our various shapes, which invokes each
        // instance's toString() method.
        // note that the actual implementation for Circle and Rectangle
        // relies on the concrete implementation in abstract class Shape,
        // which in turn relies on polymorphism to dynamically look up the
        // correct calcArea() method for the current instance's class;
        // Square provides its own standalone version
        for (Shape s: shapes)
        {
            System.out.println(s);

            // we could also call s.calcArea(), s.calcPerimeter(), s.equals(),
            // and s.toString(), as these are all Shape methods

            // note that we cannot do non-Shape things here, even though
            // the actual instances are of type Circle, Rectangle, and Square;
            // we're limited to only things all Shapes can do
            // double radius = s.getRadius();  // not all shapes have this!

            // if we are very careful, we can selectively do specific things
            if (s instanceof Circle)
            {
                // here we have found a Circle, so we can do Circle things
                // if we cast the Shape to Circle; parentheses are required
                // to get a reference to a Circle before trying to invoke
                // Circle methods such as getRadius()
                System.out.format("(Found a Circle with radius %.5f)\n",
                    ((Circle)s).getRadius()
                );

                // note that we generally don't want to do this;
                // rather, rely on substitutability and polymorphism
                // to the extent possible
            }
        }

        // finally, we can use the concrete equals() method from abstract
        // class Shape to see if one shape is equal to another, based on area
        System.out.format("Shape 1 is a %s with area %.5f and ",
            circle.getClass().getName(), circle.calcArea()
        );
        System.out.format("Shape 2 is a %s with area %.5f; they are ",
            square.getClass().getName(), square.calcArea()
        );
        if (!circle.equals(square))
        {
            System.out.print("not ");
        }
        System.out.println("equal.");

        // and another pair of shapes...
        System.out.format("Shape 2 is a %s with area %.5f and ",
            rectangle.getClass().getName(), rectangle.calcArea()
        );
        System.out.format("Shape 3 is a %s with area %.5f; they are ",
            square.getClass().getName(), square.calcArea()
        );
        if (!rectangle.equals(square))
        {
            System.out.print("not ");
        }
        System.out.println("equal.");
    }
}
